﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Reflection;

namespace InterfacesVsDelegatesApp
{
    public sealed class Program:
        IRunnable
    {
        private const int COUNT = 100000000;
        private const int COUNT2 = 10000000;

        static void Main(string[] args)
        {
            var program = new Program();
            program.Run();
        }

        private void Run()
        {
            Console.WriteLine("This application tests the speed of interfaces and delegates in");
            Console.WriteLine("different situations. Compile it in release and execute it outside");
            Console.WriteLine("Visual Studio to get the right results.");
            Console.WriteLine();

            Console.WriteLine("The following tests do 100 millions of iterations:");
            TestDelegateSpeed_WrongWay();
            TestInterfaceSpeed_WrongWay();
            Console.WriteLine();

            TestDelegateSpeed_RightWay();
            TestInterfaceSpeed_RightWay();
            Console.WriteLine();

            TestAnonymousDelegateSpeed();
            TestInterfaceWithTheSamePurposeAsAnonymousDelegateSpeed();
            Console.WriteLine();

            Console.WriteLine("The following tests do only 10 millions of iterations:");
            TestDynamicInvokeDelegateSpeed();
            TestDynamicInvokeInterfaceSpeed();
            Console.WriteLine();

            Console.WriteLine("All the tests are finished. Press ENTER to exit.");
            Console.ReadLine();
        }

        /// <summary>
        /// A method that does nothing, on purpose.
        /// </summary>
        private void SomeMethod()
        {
        }
        private void TestDelegateSpeed_WrongWay()
        {
            Console.Write("Testing delegate speed, the wrong way: ");

            Stopwatch stopwatch = new Stopwatch();
            stopwatch.Start();

            for (int i = 0; i < COUNT; i++)
            {
                Action action = SomeMethod;
                action();
            }

            stopwatch.Stop();
            Console.WriteLine(stopwatch.Elapsed);
        }
        private void TestDelegateSpeed_RightWay()
        {
            Console.Write("Testing delegate speed, the right way: ");

            Stopwatch stopwatch = new Stopwatch();
            stopwatch.Start();

            Action action = SomeMethod;
            for (int i = 0; i < COUNT; i++)
                action();

            stopwatch.Stop();
            Console.WriteLine(stopwatch.Elapsed);
        }

        void IRunnable.Run()
        {
        }
        private void TestInterfaceSpeed_WrongWay()
        {
            Console.Write("Testing interface speed, the wrong way: ");

            Stopwatch stopwatch = new Stopwatch();
            stopwatch.Start();

            for (int i = 0; i < COUNT; i++)
            {
                IRunnable runnable = this;
                runnable.Run();
            }

            stopwatch.Stop();
            Console.WriteLine(stopwatch.Elapsed);
        }
        private void TestInterfaceSpeed_RightWay()
        {
            Console.Write("Testing interface speed, the right way: ");

            Stopwatch stopwatch = new Stopwatch();
            stopwatch.Start();

            IRunnable runnable = this;
            for (int i = 0; i < COUNT; i++)
                runnable.Run();

            stopwatch.Stop();
            Console.WriteLine(stopwatch.Elapsed);
        }


        private void TestAnonymousDelegateSpeed()
        {
            Console.Write("Testing anonymous delegate speed: ");

            Stopwatch stopwatch = new Stopwatch();
            stopwatch.Start();

            for (int i = 0; i < COUNT; i++)
                MethodThatReceivesADelegate(() => SomeCall(i));

            stopwatch.Stop();
            Console.WriteLine(stopwatch.Elapsed);
        }
        private void MethodThatReceivesADelegate(Action action)
        {
            action();
        }
        private void SomeCall(int value)
        {
        }

        private void TestInterfaceWithTheSamePurposeAsAnonymousDelegateSpeed()
        {
            Console.Write("Testing an interface that does the same: ");

            Stopwatch stopwatch = new Stopwatch();
            stopwatch.Start();

            for (int i = 0; i < COUNT; i++)
            {
                ObjectThatHoldsI obj = new ObjectThatHoldsI();
                obj._program = this;
                obj._i = i;

                MethodThatReceivesADelegateLikeInterface(obj);
            }

            stopwatch.Stop();
            Console.WriteLine(stopwatch.Elapsed);
        }

        private sealed class ObjectThatHoldsI:
            IRunnable
        {
            internal Program _program;
            internal int _i;

            public void Run()
            {
                _program.SomeCall(_i);
            }
        }

        private void MethodThatReceivesADelegateLikeInterface(IRunnable runnable)
        {
            runnable.Run();
        }


        private void TestDynamicInvokeDelegateSpeed()
        {
            Console.Write("Testing delegate's DynamicInvoke speed: ");

            Stopwatch stopwatch = new Stopwatch();
            stopwatch.Start();

            Action<int> action = SomeCall;
            for (int i = 0; i < COUNT2; i++)
                action.DynamicInvoke(i);

            stopwatch.Stop();
            Console.WriteLine(stopwatch.Elapsed);
        }
        private void TestDynamicInvokeInterfaceSpeed()
        {
            Console.Write("Testing interface's DynamicInvoke speed: ");

            Stopwatch stopwatch = new Stopwatch();
            stopwatch.Start();

            IDynamicInvokable action = new DynamicInvokable();
            for (int i = 0; i < COUNT2; i++)
                action.DynamicInvoke(i);

            stopwatch.Stop();
            Console.WriteLine(stopwatch.Elapsed);
        }
        private sealed class DynamicInvokable:
            IAction<int>
        {
            public void Invoke(int parameter)
            {
                // this is equivalent to the SomeCall method. It does nothing, but
                // if you want to check, it is being called by the DynamicInvoke method.
            }

            public object DynamicInvoke(params object[] parameters)
            {
                // here I am doing all the parameters validation to make it
                // do effectively the same as the Delegate.DynamicInvoke method does.
                if (parameters == null)
                    throw new ArgumentNullException("parameters");

                if (parameters.Length != 1)
                    throw new ArgumentException("parameters length must be 1.", "parameters");

                int value = (int)parameters[0];

                // even if our invoke does nothing, I am putting a try catch to
                // really simulate the fact that an exception throws a TargetInvocationException.
                try
                {
                    Invoke(value);
                }
                catch(Exception exception)
                {
                    throw new TargetInvocationException(exception);
                }

                return null;
            }
        }
    }
}
